/*
 *  Copyright 2010 argius
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
package net.argius.stew;

import java.io.*;
import java.net.*;
import java.sql.*;
import java.util.*;

import net.argius.logging.*;

/**
 * RlN^B
 * JDBCRlNV̐ڑB
 */
public final class Connector {

    private static final Logger log = LoggerFactory.getLogger(Connector.class);

    private final String id;
    private final Properties props;
    private final Password password;

    /**
     * Connector̐B
     * @param id ID
     * @param props vpeB
     */
    public Connector(String id, Properties props) {
        assert id != null;
        if (!id.matches("^[A-Za-z0-9]*$")) {
            throw new IllegalArgumentException("illegal id : " + id);
        }
        this.id = id;
        this.props = props;
        this.password = createPasswordInstance(props.getProperty("password.class"));
        password.setTransformedString(props.getProperty("password"));
    }

    /**
     * PasswordCX^X̐B
     * @param className NX
     * @return CX^X
     */
    private static Password createPasswordInstance(String className) {
        try {
            return (Password)DynamicLoader.newInstance(className);
        } catch (Exception ex) {
            log.warn("", ex);
        }
        return new PlainTextPassword();
    }

    /**
     * Connector̐(Rs[RXgN^)B
     * @param id ID
     * @param src Rs[̃CX^X
     */
    public Connector(String id, Connector src) {
        this(id, new Properties(src.props));
    }

    /**
     * ID̎擾B
     * @return ID
     */
    public String getId() {
        return id;
    }

    /**
     * ̂̎擾B
     * @return 
     */
    public String getName() {
        return props.getProperty("name");
    }

    /**
     * classpath̎擾B
     * @return classpath
     */
    public String getClasspath() {
        return props.getProperty("classpath", "");
    }

    /**
     * driver̎擾B
     * @return driver
     */
    public String getDriver() {
        final String driver = props.getProperty("driver");
        if (log.isDebugEnabled()) {
            log.debug(String.format("driver=[%s]", driver));
        }
        return driver;
    }

    /**
     * url̎擾B
     * @return url
     */
    public String getUrl() {
        return props.getProperty("url");
    }

    /**
     * user̎擾B
     * @return user
     */
    public String getUser() {
        return props.getProperty("user");
    }

    /**
     * pX[hIuWFNg̎擾B
     * @return pX[hIuWFNg
     */
    public Password getPassword() {
        return password;
    }

    /**
     * readonly̎擾B
     * @return readonly Ȃ <code>true</code>
     */
    public boolean isReadOnly() {
        String s = props.getProperty("readonly");
        return Boolean.valueOf(s).booleanValue();
    }

    /**
     * rollback̎擾B
     * @return [obN@\gpȂ <code>true</code>
     */
    public boolean usesAutoRollback() {
        String s = props.getProperty("rollback");
        return Boolean.valueOf(s).booleanValue();
    }

    /**
     * ConnectorProperties֕ϊB
     * @return Properties
     */
    public Properties toProperties() {
        return (Properties)props.clone();
    }

    /**
     * RlNV̎擾B
     * @return RlNV
     * @throws SQLException caANZXG[ꍇ
     */
    public Connection getConnection() throws SQLException {
        Driver driver = loadDriver(getDriver(), getClasspath());
        Properties p = new Properties();
        p.setProperty("user", getUser());
        p.setProperty("password", getPassword().getRowString());
        if (!driver.acceptsURL(getUrl())) {
            throw new SQLException("invalid url : " + getUrl());
        }
        if (log.isInfoEnabled()) {
            log.info("driver.connect start");
        }
        Connection conn = driver.connect(getUrl(), p);
        if (log.isInfoEnabled()) {
            log.info("driver.connect end");
        }
        if (conn == null) {
            throw new IllegalStateException("driver returned null");
        }
        return conn;
    }

    /**
     * JDBChCo[hB
     * @param className SCNX
     * @param classpath NXpX
     * @return JDBChCo
     * @throws SQLException 
     */
    private Driver loadDriver(String className, String classpath) throws SQLException {
        try {
            String[] paths = classpath.split(File.pathSeparator);
            List<URL> list = new ArrayList<URL>();
            for (int i = 0; i < paths.length; i++) {
                try {
                    list.add(new File(paths[i]).toURL());
                } catch (MalformedURLException ex) {
                    // ignore
                }
            }
            URL[] urls = list.toArray(new URL[list.size()]);
            if (log.isInfoEnabled()) {
                log.info("driver.instantiate start");
            }
            if (this != null) {
                Object driver = DynamicLoader.newInstance(className, urls);
                if (log.isInfoEnabled()) {
                    log.info("driver.instantiate end");
                }
                return (Driver)driver;
            }
            try {
                return DriverManager.getDriver(getUrl());
            } catch (Exception ex) {
                log.warn(ex);
                ex.printStackTrace(System.out);
                try {
                    Class.forName(className,
                                  false,
                                  new URLClassLoader(urls, ClassLoader.getSystemClassLoader()));
                } catch (ClassNotFoundException ex1) {
                    throw new RuntimeException(ex.toString());
                }
                return DriverManager.getDriver(getUrl());
            }
        } catch (RuntimeException ex) {
            log.error("", ex);
            SQLException sqlError = new SQLException("failed to load driver");
            sqlError.initCause(ex);
            throw sqlError;
        }
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((props == null) ? 0 : props.hashCode());
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (!(obj instanceof Connector)) {
            return false;
        }
        Connector other = (Connector)obj;
        return props.equals(other.props);
    }

    @Override
    public String toString() {
        return "Connector:" + id;
    }

}